-- Spatial Prisioner's Dillema
--[[
The evolutionary puzzle of cooperation describes situations 
where cooperators provide a fitness benefit to other individuals at some cost to themselves.  
One well-studied possibility to maintain cooperation is to consider structured populations, 
where each individual interacts only with a limited subset of the population. 
This enables cooperators to form clusters such that they are more likely to interact with 
other cooperators instead of being exploited by defectors. Here we present a detailed analysis 
of how a few cooperators invade and expand in a world of defectors. If the invasion succeeds, 
the expansion process takes place in two stages: first, cooperators and defectors quickly establish a 
local equilibrium and then they uniformly expand in space. The second stage provides good estimates 
for the global equilibrium frequencies of cooperators and defectors. Under hospitable conditions, 
cooperators typically form a single, ever growing cluster interspersed with specks of defectors, 
whereas under more hostile conditions, cooperators form isolated, compact clusters that minimize 
exploitation by defectors. We provide the first quantitative assessment of the way cooperators arrange 
in space during invasion and find that the macroscopic properties and the emerging spatial patterns 
reveal information about the characteristics of the underlying microscopic interactions.

]]
-- 
-- Under what conditions will there be a spatial coexistence of cooperators and defectors, and
-- When will we observe the extinction of either cooperators or defectors?

-- What does the variety and the size of the spatial domains depend on?

-- Will the dynamics always lead to stationary patterns, or will there be also non-stationary
--    patterns in the long run?

-- How does the dynamics changes if a larger local neighborhood is considered in the update
--    rule (i.e. if the second nearest neighbors are included, too)?

-- The following example uses the following rules:

-- Cooperators can survive by forming clusters and thereby outweighing losses against defectors.
-- There is a balance between payoff and survival rates

-- The following example uses the following rules:

-- 1. Each player occupies a single cell
-- 2. Players compete against  4 neighbors in an 3 x 3 cell 
-- 4. Updates are synchronous 
-- 5. Each cell copies the most successful strategy in its neighborhood

--[[  Paper references 
Philipp Langera, Martin A. Nowak, Christoph Hauerta, "Spatial invasion of cooperation". 
Journal of Theoretical Biology 250 (2008) 634–641

]] 



COOPERATE = 0
DEFECT    = 1
COOPERATE_TO_DEFECT = 2
DEFECT_TO_COOPERATE = 3

-- Payoffs
-- (T)emptation > (R)eward > (P)unishment > (S)ucker's payoff
-- T > R > P > S is the condition for prisioner's dillema


TEMPTATION = 1.5
REWARD     = 1
PUNISHMENT = 0.2
SUCKER     = 0

CLUSTER_SIZE = 15

function createCellSpace (model)
	model.cellspace = CellularSpace {
		xdim = model.cellSpaceDim,
		ydim = model.cellSpaceDim
	}

    model.cellspace:createNeighborhood {
        strategy = "moore",
  		self = true,
  		wrap = true
	}

end

function createAgent (model)

	model.player = Agent {
	-- an Agent has two functions (init and execute)
	-- init initializes the agent
	-- execute runs the agent
		-- initialize the agent

		init = function (self)
				self.strategy = DEFECT
		end,

		execute = function (self)
                --- em branco
		end
	}
end

function createSociety (model)
	model.society = Society {
		instance = model.player, 
		quantity = model.societyDim
	}
end

function createEnvironment (model)

	model.env = Environment {model.cellspace, model.society}

	model.env:createPlacement {
		strategy = "random",
		max = 1 
	}

	forEachAgent (model.society, function (agent)
		agent:getCell().state = agent.strategy
	end)

	midpoint = (model.cellSpaceDim - 1)/2

	box = {
		upperleft  = midpoint - (CLUSTER_SIZE - 1)/2,
		lowerright = midpoint + (CLUSTER_SIZE - 1)/2 
	}

	model.cluster = Trajectory{ 
		target = model.cellspace, 
		select = function (cell)
			return (cell.x > box.upperleft and cell.x < box.lowerright ) and
			       (cell.y > box.upperleft and cell.y < box.lowerright )
		end
	}

	forEachCell (model.cluster, function (cell)
		print ("x y ", cell.x, cell.y)
		cell:getAgent().strategy = COOPERATE
		cell.state = COOPERATE
	end)

end

function createTimer (model)
	model.timer = Timer {
		Event {time = 1, period = 1, action = function (e)
			forEachAgent (model.society, function (agent)
					playWithNeighbors (agent)
				end)
			forEachAgent (model.society, function(agent)
					chooseBestStrategy (agent)
				end)
			forEachAgent (model.society, function(agent)
					changeStrategy (agent)
				end)
			stats (model)
			model.cellspace:notify()
			-- model.plot:notify(e:getTime())
			end}
		}
end

function playGame (agent1, agent2)
	if agent1.strategy == COOPERATE and agent2.strategy == COOPERATE then
		return REWARD
	elseif agent1.strategy == COOPERATE and agent2.strategy == DEFECT then
		return SUCKER
	elseif agent1.strategy == DEFECT    and agent2.strategy == COOPERATE then
		return TEMPTATION
	elseif agent1.strategy == DEFECT    and agent2.strategy == DEFECT then 
		return PUNISHMENT
	end
end

function playWithNeighbors (agent)
	cell = agent:getCell()
	agent.payoff = 0 
	forEachNeighbor (cell, function (cell, neigh)
			other = neigh:getAgent()
			result = playGame (agent, other)
			agent.payoff = agent.payoff + result
		end)
end

function chooseBestStrategy (agent)
	cell = agent:getCell()
	agent.beststrategy = agent.strategy
	bestpayoff = agent.payoff
	
	forEachNeighbor (cell, function (cell, neigh)
			other = neigh:getAgent()
			if (other.payoff > bestpayoff) then
				bestpayoff   = other.payoff
				agent.beststrategy = other.strategy
			end
	end)

end

function changeStrategy (agent)
	
	if (agent.strategy == COOPERATE) then
		if (agent.beststrategy == DEFECT) then
			agent.strategy        = DEFECT
			agent:getCell().state = COOPERATE_TO_DEFECT
		else 
			agent:getCell().state = COOPERATE
		end
	else  -- agent.strategy == DEFECT 
		if (agent.beststrategy == COOPERATE) then
			agent.strategy        = COOPERATE
			agent:getCell().state = DEFECT_TO_COOPERATE
		else
			agent:getCell().state = DEFECT
		end
	end 

end

function stats (model)
	local num_coop = 0 
	
	forEachCell (model.cellspace, function (cell) 
		if cell:getAgent().strategy == COOPERATE  then 
			num_coop = num_coop + 1 
		end
	end)
	model.variables.percent_cooperators = (100*num_coop)/((model.cellSpaceDim)*(model.cellSpaceDim))
	
end

function createMap (model)

	leg = Legend {
		grouping = "uniquevalue",
		colorBar = {
			{value = COOPERATE, color = "blue"},
			{value = DEFECT,    color = "red"},
			{value = COOPERATE_TO_DEFECT,   color = "yellow"},
			{value = DEFECT_TO_COOPERATE,   color = "green"}
		}
	}
	map = Observer {
		type = "map",
		subject = model.cellspace,
		attributes = {"state"},
		legends = {leg}
	}
end

function createObserverChart (terrame_object, plotLabels)
	chart = Observer {
		subject     = terrame_object,
		attributes  = plotLabels.attributes,
		curveLabels = plotLabels.attributes,
		title       = plotLabels.title,
		xLabel      = plotLabels.xLabel,
		yLabel      = plotLabels.yLabel,
		type = "chart"
	}
end

function createPlot (model)

	model.plot = Cell (model.variables)
	createObserverChart (model.plot, model.plotLabels)
	model.plot:notify(0) 

end

function AgentModel (model)

	math.randomseed(os.time())

	-- createCellSpace 
	createCellSpace (model)  -- Ok

	-- create Agent
	createAgent (model)   -- OK

	-- create Society
	createSociety (model) --oK

	-- create Environment 
	createEnvironment (model)

	-- timer 
	createTimer (model)

	-- create Observer Map
	createMap (model)

	-- create Observer Chart
	-- createPlot (model)

	return model
end

function run (model, time)
	model.timer:execute (time)
end

model = AgentModel {
	cellSpaceDim  =  71,
	societyDim    =  5041,
	variables     = { percent_cooperators = 0 },
	plotLabels    = {
		attributes = {"percent_cooperators"},
		title      = {"% of Cooperators"},
		xLabel     = {"time"},
		yLabel     = {"Percent"}
	}
}

run (model, 300)